/*
 * $Id: presentity.c 2288 2007-05-24 16:42:35Z anca_vamanu $
 *
 * presence module - presence server implementation
 *
 * Copyright (C) 2006 Voice Sistem S.R.L.
 *
 * This file is part of openser, a free SIP server.
 *
 * openser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * openser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * History:
 * --------
 *  2006-08-15  initial version (anca)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <libxml/parser.h>

#include "../../db/db.h"
#include "../../dprint.h"
#include "../../mem/shm_mem.h"
#include "../../str.h"
#include "../alias_db/alias_db.h"
#include "../../data_lump_rpl.h"
#include "pidf.h"
#include "presentity.h"
#include "presence.h" 
#include "notify.h"

extern int use_db;
extern char* presentity_table;
extern db_con_t* pa_db;
extern db_func_t pa_dbf;

static str pu_200_rpl  = str_init("OK");
static str pu_412_rpl  = str_init("Conditional request failed");

int publ_send200ok(struct sip_msg *msg, int lexpire, str etag)
{
	str hdr_append, hdr_append2 ;

	DBG("PRESENCE:publ_send200ok: send 200OK reply\n");	
	DBG("PRESENCE:publ_send200ok: etag= %s - len= %d\n", etag.s, etag.len);
	/* ??? should we use static allocated buffer */
	hdr_append.s = (char *)pkg_malloc( sizeof(char)*100);
	if(hdr_append.s == NULL)
	{
		LOG(L_ERR,"ERROR:publ_send200ok: unable to add lump_rl\n");
		return -1;
	}
	hdr_append.s[0]='\0';
	hdr_append.len = sprintf(hdr_append.s, "Expires: %d\r\n",lexpire -
			expires_offset);
	if(hdr_append.len < 0)
	{
		LOG(L_ERR, "PRESENCE:publ_send200ok: ERROR unsuccessful sprintf\n");
		pkg_free(hdr_append.s);
			return -1;
	}
	hdr_append.s[hdr_append.len]= '\0';
		
	if (add_lump_rpl( msg, hdr_append.s, hdr_append.len, LUMP_RPL_HDR)==0 )
	{
		LOG(L_ERR,"PRESENCE: publ_send200ok:ERROR unable to add lump_rl\n");
		pkg_free(hdr_append.s);
		return -1;
	}
	pkg_free(hdr_append.s);

	hdr_append2.s = (char *)pkg_malloc( sizeof(char)*(16+etag.len) );
	if(hdr_append2.s == NULL)
	{
		LOG(L_ERR,"PRESENCE:publ_send200ok:ERROR unable to add lump_rl\n");
		return -1;
	}
	hdr_append2.s[0]='\0';
	hdr_append2.len = sprintf(hdr_append2.s, "SIP-ETag: %s\r\n", etag.s);
	if(hdr_append2.len < 0)
	{
		LOG(L_ERR, "PRESENCE:publ_send200ok:ERROR unsuccessful sprintf\n ");
		pkg_free(hdr_append2.s);
		return -1;
	}
	hdr_append2.s[hdr_append2.len]= '\0';
	if (add_lump_rpl(msg, hdr_append2.s, hdr_append2.len, LUMP_RPL_HDR)==0 )
	{
		LOG(L_ERR,"PRESENCE:publ_send200ok: unable to add lump_rl\n");
		pkg_free(hdr_append2.s);
		return -1;
	}
	pkg_free(hdr_append2.s);

	if( slb.reply( msg, 200, &pu_200_rpl)== -1)
	{
		LOG(L_ERR,"PRESENCE: publ_send200ok: ERORR while sending reply\n");
		return -1;
	}

	return 0;

}	

int new_presentity( str* domain,str* user,
		int expires, int received_time, str* etag, presentity_t **p)
{
	presentity_t *presentity;
	int size;
	
	/* alocating memory for presentity */
	size = sizeof(presentity_t)+ domain->len+ user->len+ etag->len + 1;
	presentity = (presentity_t*)pkg_malloc(size);
	if(presentity == NULL)
	{
		LOG(L_ERR, "PRESENCE:new_presentity: No memory left: size=%d\n", size);
		return -1;
	}
	memset(presentity, 0, sizeof(presentity_t));

    
	presentity->domain.s = ((char*)presentity)+sizeof(presentity_t);
	strncpy(presentity->domain.s, domain->s, domain->len);
	presentity->domain.s[domain->len] = 0;
	presentity->domain.len = domain->len;
	
	presentity->user.s = ((char*)presentity)+sizeof(presentity_t)+domain->len;
	strncpy(presentity->user.s, user->s, user->len);
	presentity->user.s[user->len] = 0;
	presentity->user.len = user->len;

	presentity->etag.s = ((char*)presentity)+sizeof(presentity_t)
							+domain->len+ user->len;
	strncpy(presentity->etag.s, etag->s, etag->len);
	presentity->etag.s[etag->len] = 0;
	presentity->etag.len = etag->len;

	presentity->expires = expires;
	presentity->received_time = received_time;

	*p=presentity;

	return 0;
}

int update_presentity(struct sip_msg* msg, presentity_t* presentity,
		str* body, int new_t )
{
	if( !use_db )
		return 0;

	db_key_t query_cols[7];
	db_op_t  query_ops[7];
	db_val_t query_vals[7], update_vals[3];
	db_key_t result_cols[4], update_keys[3];
	db_res_t *result= NULL;

	int n_query_cols = 0;
	int n_result_cols = 0;
	int n_update_cols = 0;
	int body_col;
    
	query_cols[0] = "domain";
	query_ops[0] = OP_EQ;
	query_vals[0].type = DB_STR;
	query_vals[0].nul = 0;
	query_vals[0].val.str_val.s = presentity->domain.s;
	query_vals[0].val.str_val.len = presentity->domain.len;

	n_query_cols++;
	
	query_cols[1] = "username";
	query_ops[1] = OP_EQ;
	query_vals[1].type = DB_STR;
	query_vals[1].nul = 0;
	query_vals[1].val.str_val.s = presentity->user.s;
	query_vals[1].val.str_val.len = presentity->user.len;
	n_query_cols++;

	
	query_cols[n_query_cols] = "etag";
	query_ops[n_query_cols] = OP_EQ;
	query_vals[n_query_cols].type = DB_STR;
	query_vals[n_query_cols].nul = 0;
	query_vals[n_query_cols].val.str_val.s = presentity->etag.s;
	query_vals[n_query_cols].val.str_val.len = presentity->etag.len;
	n_query_cols++;
			

	if(presentity->expires == 0) 
	{	
		if( publ_send200ok(msg, presentity->expires, presentity->etag)< 0)
		{
			LOG(L_ERR, "PRESENCE:update_presentity: ERROR while sending 200OK\n");
			return -1;
		}

		query_db_notify( &presentity->user, &presentity->domain, 
					"presence", NULL, &presentity->etag);
			
		if (pa_dbf.use_table(pa_db, presentity_table) < 0) 
		{
			LOG(L_ERR, "PRESENCE:update_presentity: Error in use_table\n");
			goto error;
		}
		DBG("PRESENCE:update_presentity: expires =0 -> deleting"
				" from database\n");
		if(pa_dbf.delete(pa_db, query_cols, 0 ,query_vals,n_query_cols)< 0 )
		{
			DBG( "PRESENCE:update_presentity: ERROR cleaning"
					" unsubscribed messages\n");
		}
		DBG("PRESENCE:update_presentity:delete from db %.*s\n",
				presentity->user.len,presentity->user.s );
		return 1;
	}

	if(new_t) /* daca a fost generat un nou etag insereaza */
	{
		xmlDocPtr doc = NULL;
		xmlNodePtr root_node = NULL;
		char* status = NULL;

		/* insert new record into database */	
		query_cols[n_query_cols] = "event";
		query_vals[n_query_cols].type = DB_STR;
		query_vals[n_query_cols].nul = 0;
		query_vals[n_query_cols].val.str_val.s = "presence";
		query_vals[n_query_cols].val.str_val.len= strlen("presence");
		n_query_cols++;
			
		query_cols[n_query_cols] = "expires";
		query_vals[n_query_cols].type = DB_INT;
		query_vals[n_query_cols].nul = 0;
		query_vals[n_query_cols].val.int_val = presentity->expires+
				(int)time(NULL);
		n_query_cols++;
	
		query_cols[n_query_cols] = "received_time";
		query_vals[n_query_cols].type = DB_INT;
		query_vals[n_query_cols].nul = 0;
		query_vals[n_query_cols].val.int_val = presentity->received_time;
		n_query_cols++;

		query_cols[n_query_cols] = "body";
		query_vals[n_query_cols].type = DB_BLOB;
		query_vals[n_query_cols].nul = 0;
		query_vals[n_query_cols].val.str_val.s = body->s;
		query_vals[n_query_cols].val.str_val.len = body->len;
		n_query_cols++;
			
		if (pa_dbf.use_table(pa_db, presentity_table) < 0) 
		{
			LOG(L_ERR, "PRESENCE:update_presentity: Error in use_table\n");
			goto error;
		}

		DBG( "PRESENCE:update_presentity: inserting %d cols into"
				"table\n",
				n_query_cols);
				
		if (pa_dbf.insert(pa_db, query_cols, query_vals, n_query_cols) < 0) 
		{
			LOG(L_ERR, "PRESENCE:update_presentity: Error while"
					" inserting new presentity\n");
			goto error;
		}
				/* send 200ok */
		if(publ_send200ok(msg, presentity->expires, presentity->etag)< 0)
		{
			LOG(L_ERR, "PRESENCE:update_presentity: ERROR while sending 2000k\n");
			goto error;
		}

		doc = xmlParseMemory(body->s, body->len);
		if ( doc == NULL)
		{
			LOG(L_ERR, "PRESENCE:update_presentity: ERROR while parsing"
					" xml body\n");
			goto error;
		}
		root_node = xmlDocGetNodeByName(doc,"presence", NULL);
		if(root_node == NULL)
		{
			LOG(L_ERR, "PRESENCE:update_presentity: ERROR while getting"
					" entity attribute\n");
			goto error;
		}

		status = xmlNodeGetNodeContentByName(root_node, "basic", NULL );
	
		if(status == NULL)
		{
			LOG(L_ERR, "PRESENCE:update_presentity: ERROR while getting" 
					" entity attribute\n");
			xmlFreeDoc(doc);
			xmlCleanupParser();
			goto error;
		}

		if(strncmp (status,"closed", 6) == 0)
		{
			DBG("PRESENCE:update_presentity:The presentity status" 
					" is offline; do not send notify\n");
		}
		else			/* send notify with presence information */
			if (query_db_notify(&presentity->user, &presentity->domain,
						"presence", NULL, NULL)<0)
			{
				DBG(" PRESENCE:update_presentity:Could not send"
						" notify for event presence\n");
			//	goto error;
			}
		xmlFree(status);
		xmlFreeDoc(doc);
		xmlCleanupParser();
	}
	else
	{
		result_cols[body_col=n_result_cols++] = "expires" ;
		if (pa_dbf.use_table(pa_db, presentity_table) < 0) 
		{
			LOG(L_ERR, "PRESENCE:update_presentity: Error in use_table\n");
			goto error;
		}

		if (pa_dbf.query (pa_db, query_cols, query_ops, query_vals,
			 result_cols, 3, n_result_cols, 0, &result) < 0) 
		{
			LOG(L_ERR, "PRESENCE:update_presentity: Error while querying"
					" presentity\n");
			goto error;
		}
		if(result== NULL)
			goto error;

		if (result->n > 0)
		{
			pa_dbf.free_result(pa_db, result);
			result= NULL;
			
			if(body==NULL || body->s==NULL) /* if there is no body update expires value */
			{
				update_keys[0] = "expires";
				update_vals[0].type = DB_INT;
				update_vals[0].nul = 0;
				update_vals[0].val.int_val = presentity->expires + (int)time(NULL);

				update_keys[1] = "received_time";
				update_vals[1].type = DB_INT;
				update_vals[1].nul = 0;
				update_vals[1].val.int_val = presentity->received_time;


				if( pa_dbf.update( pa_db,query_cols, query_ops,
				query_vals, update_keys, update_vals, n_query_cols,2 )<0) 
				{
					LOG( L_ERR , "PRESENCE:update_presentity:ERROR while"
							" updating presence information\n");
					goto error;
				}
				
				/* send 200ok */
				if(publ_send200ok(msg, presentity->expires, presentity->etag)< 0)
				{
					LOG(L_ERR, "PRESENCE:update_presentity: ERROR while sending 2000k\n");
					goto error;
				}

				return 0;
			}

			/* write the new body*/
			update_keys[0] = "body";
			update_vals[0].type = DB_BLOB;
			update_vals[0].nul = 0;
			update_vals[0].val.str_val.s = body->s;
			update_vals[0].val.str_val.len=body->len;
			n_update_cols++;
			
			update_keys[n_update_cols] = "expires";
			update_vals[n_update_cols].type = DB_INT;
			update_vals[n_update_cols].nul = 0;
			update_vals[n_update_cols].val.int_val = presentity->expires+ (int)time(NULL);
			n_update_cols++;

			update_keys[n_update_cols] = "received_time";
			update_vals[n_update_cols].type = DB_INT;
			update_vals[n_update_cols].nul = 0;
			update_vals[n_update_cols].val.int_val = presentity->received_time;
			n_update_cols++;


			if( pa_dbf.update( pa_db,query_cols, query_ops, query_vals,
			update_keys,update_vals, n_query_cols, n_update_cols )<0) 
			{
				LOG( L_ERR , "PRESENCE:update_presentity: ERROR while"
						" updating presence information\n");
				goto error;
			}

			/* send 200OK */
			if( publ_send200ok(msg, presentity->expires,presentity->etag)< 0)
			{
				LOG(L_ERR, "PRESENCE:update_presentity: ERROR while sending 200OK\n");
				return -1;
			}

			/* presentity body is updated so send notify to all watchers */
			if (query_db_notify(&presentity->user, &presentity->domain,
						"presence", NULL, NULL)<0)
			{
				LOG(L_ERR," PRESENCE:update_presentity: Could not send Notify\n");
		//		goto error;
			}
		}  
		else  /* if there isn't no registration with those 3 values */
		{
			pa_dbf.free_result(pa_db, result);
			result= NULL;
			LOG(L_DBG, "PRESENCE:update_presentity: No E_Tag match\n");
			if (slb.reply(msg, 412, &pu_412_rpl) == -1)
			{
				LOG(L_ERR, "PRESENCE:PRESENCE:update_presentity: ERROR while sending"
					"reply\n");
				goto error;
			}
		}
	
	}
	
	return 0;

error:
	LOG(L_ERR, "PRESENCE:update_presentity: ERROR occured\n");
	if(result)
	{
		pa_dbf.free_result(pa_db, result);
		result= NULL;
	}
	return -1;

}

void free_presentity(presentity_t* presentity)
{
	pkg_free(presentity);
}


